"""Constants for alert metrics."""

# ALERT RELATED CONSTANTS #
# The valid stats commonly calculated for ACE Alerts
VALID_ALERT_STATS = ['cycle_time_sum',
                     'cycle_time_mean',
                     'cycle_time_min',
                     'cycle_time_max',
                     'cycle_time_std',
                     'alert_count'
                    ]

# for mapping stats to human readable descriptions
FRIENDLY_STAT_NAME_MAP = {
    'cycle_time_sum': "Total Open Time",
    'cycle_time_mean': "Average Time to Disposition",
    'cycle_time_min': "Quickest Disposition",
    'cycle_time_max': "Slowest Disposition",
    'cycle_time_std': "Standard Deviation for Time to Disposition",
    'alert_count': "Alert Quantities by Disposition"
}

# The core ACE specific killchain dispositions.
# The order of these dispositions should NOT change.
# If you add more, add them to the end of the list.
# Also, NOTE that any additional dispositions found in 
# ACE alert data will be alphabetically organized and then
#  extended to this list.
ACE_KILLCHAIN_DISPOSITIONS = [ 'FALSE_POSITIVE',
                               'GRAYWARE',
                               'POLICY_VIOLATION',
                               'RECONNAISSANCE',
                               'WEAPONIZATION',
                               'DELIVERY',
                               'EXPLOITATION',
                               'INSTALLATION',
                               'COMMAND_AND_CONTROL',
                               'EXFIL',
                               'DAMAGE'
                               ]

# Database query for getting alerts between two dates.
# Allows for reduction by list of company id
ALERTS_BY_MONTH_DB_QUERY = """SELECT DATE_FORMAT(insert_date, '%%Y%%m') AS month, insert_date, disposition,
                              disposition_time, disposition_user_id, owner_id, owner_time FROM alerts
                              WHERE insert_date BETWEEN %s AND %s AND alert_type!='faqueue'
                              AND alert_type!='dlp - internal threat' AND alert_type!='dlp-exit-alert' 
                              AND disposition IS NOT NULL {}{}"""

# Database query for getting alerts between two dates
# by alert type.
# Allows for reduction by list of company id
ALERTS_BY_ALERT_TYPE_QUERY = """select DATE_FORMAT(insert_date, '%%Y%%m') AS month, insert_date, disposition, disposition_time,
                                disposition_user_id, event_time
                                FROM alerts WHERE alert_type=%s
                                AND insert_date BETWEEN %s AND %s AND disposition is not NULL {}{}
                             """


# Database query for counting alerts by month and user
ALERTS_BY_MONTH_AND_USER_QUERY = """SELECT count(a.id) AS 'Alerts', DATE_FORMAT(a.insert_date,'%%Y%%m') AS Month, users.username AS User
                                    FROM alerts a JOIN users ON users.id=a.disposition_user_id WHERE a.alert_type!='faqueue'
                                    AND alert_type!='dlp - internal threat' AND alert_type!='dlp-exit-alert' AND
                                    DATE_FORMAT(a.insert_date, '%%Y%%m')=%s and users.username=%s GROUP BY Month
                                  """

# Database query for counting alerts by month and user
ALERTS_BY_MONTH_AND_TYPE_QUERY = """SELECT count(id) AS 'Alerts', DATE_FORMAT(insert_date,'%%Y%%m') AS Month
                                    FROM alerts WHERE DATE_FORMAT(insert_date, '%%Y%%m')=%s
                                    AND ( {} ) GROUP BY Month
                                  """



"""Metric functions for ACE Alert database data.

Every function in this file should expliclty work with
data from the ACE `ace.alerts` database table for the
purpose of generating metrics.

NOTE: An argument could be made that some of these
  functions, like the business time helpers, could
  move to `metrics.helpers` but so far they're only
  needed here.
"""

import os
import logging

import pymysql
import pytz
import businesstime

import pandas as pd

from typing import Tuple, Mapping, Optional
from datetime import timedelta, datetime, time
from dateutil.relativedelta import relativedelta
from businesstime.holidays import Holidays

from .constants import VALID_ALERT_STATS, FRIENDLY_STAT_NAME_MAP, ACE_KILLCHAIN_DISPOSITIONS, ALERTS_BY_MONTH_DB_QUERY


class SiteHolidays(Holidays):
    rules = [
        dict(name="New Year's Day", month=1, day=1),
        #dict(name="Birthday of Martin Luther King, Jr.", month=1, weekday=0, week=3),
        #dict(name="Washington's Birthday", month=2, weekday=0, week=3),
        dict(name="Memorial Day", month=5, weekday=0, week=-1),
        dict(name="Independence Day", month=7, day=4),
        dict(name="Labor Day", month=9, weekday=0, week=1),
        #dict(name="Columbus Day", month=10, weekday=0, week=2),
        #dict(name="Veterans Day", month=11, day=11),
        dict(name="Thanksgiving Day", month=11, weekday=3, week=4),
        dict(name="Day After Thanksgiving Day", month=11, weekday=4, week=4),
        dict(name="Chistmas Eve", month=12, day=24),
        dict(name="Chistmas Day", month=12, day=25),
    ]

    def _day_rule_matches(self, rule, dt):
        """
        Day-of-month-specific US federal holidays that fall on Sat or Sun are
        observed on Fri or Mon respectively. Note that this method considers
        both the actual holiday and the day of observance to be holidays.
        """
        if dt.weekday() == 4:
            sat = dt + timedelta(days=1)
            if super(SiteHolidays, self)._day_rule_matches(rule, sat):
                return True
        elif dt.weekday() == 0:
            sun = dt - timedelta(days=1)
            if super(SiteHolidays, self)._day_rule_matches(rule, sun):
                return True
        return super(SiteHolidays, self)._day_rule_matches(rule, dt)

def define_business_time(start_hour=6,
                         end_hour=18,
                         time_zone='US/Eastern',
                         holidays=SiteHolidays()
                         ) -> businesstime.BusinessTime:
    """Create a businesstime.BusinessTime object as defined by the passed arguments.

    Args:
        start_hour: The business time start hour represented as in integer.
        end_hour: The business time end hour represented as an integer.
        time_zone: The time zone that should be converted to before business
          hours are applied.
        holidays: A businesstime.holidays.Holidays based object representing
          the relevant holidays to exclude from business time.
    """
    business_hours = (time(start_hour), time(end_hour))
    business_hours = businesstime.BusinessTime(business_hours=business_hours,
                                               holidays=holidays)
    # HACK: Assign dynamic properties
    business_hours.time_zone = time_zone
    business_hours._start_hour = start_hour
    business_hours._end_hour = end_hour
    return business_hours

def get_alerts_between_dates(start_date: datetime,
                             end_date: datetime,
                             con: pymysql.connections.Connection,
                             alert_query: str =ALERTS_BY_MONTH_DB_QUERY,
                             selected_companies: list =[],
                            ) -> pd.DataFrame:
    """Query the database for all ACE alerts between two dates.

    Query the ACE database for all ACE alerts betweem two dates.
    If there are selected_companies, only return alerts associated to those companies,
    else all alerts are selected.

    Args:
        start_date: Get alerts created on or after this datetime.
        end_date: Get alert created on or before this datetime.
        con: a pymysql database connectable
        alert_query: The str SQL database query to get alerts with.
        selected_companies: A list of companies to select alerts for, by name.
          If the list is empty, all alerts are selected.

    Returns:
        A pd.DataFrame of the alerts.
    """

    # apply company selection by name
    company_ids = []
    if selected_companies:
        cursor = con.cursor()
        cursor.execute("select * from company")
        for c_id,c_name in cursor.fetchall():
            if c_name in selected_companies:
                company_ids.append(c_id)

    alert_query = alert_query.format(' AND ' if company_ids else '', '( ' + ' OR '.join(['company_id=%s' for company in selected_companies]) +') ' if company_ids else '')

    params = [start_date.strftime('%Y-%m-%d %H:%M:%S'),
              end_date.strftime('%Y-%m-%d %H:%M:%S')]
    params.extend(company_ids)

    alerts = pd.read_sql_query(alert_query, con, params=params)

    # go ahead and drop the dispositions we don't care about
    # XXX make a list of dispositions to ignore configurable
    alerts = alerts[alerts.disposition != 'IGNORE']
    alerts.set_index('month', inplace=True)

    return alerts

def _datetime_to_time_zone(dt=None, time_zone='US/Eastern'):
    """Convert a datetime.datetime object to the equivalent business hour time zone.

    Args:
        dt: A datatime.datetime object to convert. If dt is None, convert the current
          UTC time to the business hour time zone.
        bh_tz: A pytz timezone that the business hours are in.

    Returns:
        The df datetime.datetime in the business hour time zone.
    """
    if dt is not None:
        assert isinstance(dt, datetime)
    else:
        dt = datetime.utcnow()

    if dt.tzinfo is None:
        dt = pytz.utc.localize(dt)

    bh_tz = pytz.timezone(time_zone)
    # convert to the business hour time zone
    dt = dt.astimezone(bh_tz)
    # XXX TODO: Revisit and see if this is still neccessary:
    # because the businesshour library's math in -> def _build_spanning_datetimes(self, d1, d2) throws
    # an error if datetime.datetime objects are time zone aware, we make the datetime naive again, 
    # however, the replace method trys to be smart and convert the time back to UTC.. so we explicitly
    # make `replace` keep the hour set to the business time zone hour.
    return dt.replace(hour=dt.hour, tzinfo=None)

def get_business_hour_cycle_time(alerts: pd.DataFrame,
                                 business_hours: Optional[businesstime.BusinessTime] = None
                                ) -> pd.Series(timedelta):
    """Convert alert times to a business hours timedelta pd.Series.

    From a DataFrame of alerts with insert_date and disposition_time, generate
    a pd.Series(timedelta) calculated by only counting time that falls into
    the defined business hours. 24 hour clock.

    Args:
        alert_time_df: A pandas DataFrame with an insert_date column
          and a disposition_time column.
        business_hours: None or the businesstime.BusinessTime representation
          of the business hours to use  for calculating time base
          statistics with business hours applied.  See `define_business_time()`.

    Returns: pd.Series(timedelta) of alert cycle times in business hours
    """

    if business_hours is None:
        # use defaults
        business_hours = define_business_time()

    bh_cycle_time = []
    logging.debug(f"calculating business hours for {len(alerts)} alerts...")
    for alert in alerts.itertuples():
        open_hours = business_hours.open_hours.seconds / 3600
        bh_insert_date = _datetime_to_time_zone(alert.insert_date, time_zone=business_hours.time_zone)
        bh_disposition_time = _datetime_to_time_zone(alert.disposition_time, time_zone=business_hours.time_zone)
        btd = business_hours.businesstimedelta(bh_insert_date, bh_disposition_time)
        btd_hours = btd.seconds / 3600
        bh_cycle_time.append(timedelta(hours=(btd.days * open_hours + btd_hours)))

    return pd.Series(data=bh_cycle_time)

def alert_statistics_by_disposition(alerts: pd.DataFrame,
                                    business_hours: Optional[businesstime.BusinessTime] = None
                                    ) -> pd.DataFrame:
    """Calculate statistics for a dataframe of alerts orgainzed by disposition.

    Given a dataframe of ACE alerts, calcuate common statistics for the subsets of 
    the alerts that are defined by disposition grouping.

    Args:
        alerts: A pandas.DataFrame of ACE alerts with the following columns:
          ['month', 'insert_date', 'disposition', 'disposition_time', 
           'disposition_user_id', 'event_time']
        business_hours: None or the businesstime.BusinessTime representation
          of the business hours to use  for calculating time base
          statistics with business hours applied.  See `define_business_time()`.

    Returns:
        A pandas.DataFrame where the rows are the calculated statistics and
        the columns are dispositions.
    """

    alerts.set_index('disposition', inplace=True)
    dispositions = alerts.index.get_level_values('disposition').unique()

    dispo_data = {}
    for dispo in dispositions:
        logging.debug(f"calculating alert statistic dispositions stats for {dispo} @ length {len(alerts)}")
        if business_hours: # could just pass df or a copy of df - here a copy with just data needed
            alert_cycle_times = get_business_hour_cycle_time(alerts.loc[[dispo],['disposition_time', 'insert_date']], business_hours)
        else:
            alert_cycle_times = alerts.loc[dispo, 'disposition_time'] - alerts.loc[dispo, 'insert_date']

        # XXX also record event-alert type (mean time to detect based on given event_time)
        #event_to_alert_time = alerts.loc[dispo, 'event_time'] - alerts.loc[dispo, 'insert_date']
            
        try:
            dispo_data[dispo] = {
                'cycle_time_sum' : alert_cycle_times.sum(),
                'cycle_time_mean' : alert_cycle_times.mean(),
                'cycle_time_min' : alert_cycle_times.min(),
                'cycle_time_max' : alert_cycle_times.max(),
                'cycle_time_std' : alert_cycle_times.std(),
                #'detection_time_sum': event_to_alert_time.sum(),
                #'detection_time_mean': event_to_alert_time.mean(),
                #'detection_time_min': event_to_alert_time.min(),
                #'detection_time_max': event_to_alert_time.max(),
                #'detection_time_std': event_to_alert_time.std(),
                'alert_count' : len(alerts.loc[dispo])
            }
        except AttributeError: # this occures when there was only ONE alert of this dispo type
            dispo_data[dispo] = {
                'cycle_time_sum' : alert_cycle_times,
                'cycle_time_mean' : alert_cycle_times,
                'cycle_time_min' : alert_cycle_times,
                'cycle_time_max' : alert_cycle_times,
                'cycle_time_std' : pd.Timedelta(timedelta()),
                #'detection_time_sum': event_to_alert_time,
                #'detection_time_mean': event_to_alert_time,
                #'detection_time_min': event_to_alert_time,
                #'detection_time_max': event_to_alert_time,
                #'detection_time_std': pd.Timedelta(timedelta()),
                'alert_count' : 1
            }

    dispo_df = pd.DataFrame(data=dispo_data, columns=dispositions)
    return dispo_df

# XXX later create a statistics by day by dispo
def statistics_by_month_by_dispo(alerts: pd.DataFrame,
                                 stats=VALID_ALERT_STATS,
                                 business_hours: Optional[businesstime.BusinessTime] = None
                                ) -> Mapping[str, pd.DataFrame]:
    """Calculate statistics for ACE alerts organized by month and disposition.

    Given a pandas.DataFrame of ACE alerts, first organize the alerts by month.
    Next, calculate statisitcs for every month of alerts by disposition.
    Finally, organize all of the statistics by statistic type pointing to a 
    DataFrame of calculated month x disposition statistics.

    Args:
        alerts: A pandas.DataFrame of ACE alerts with these columns:
            ['month', 'insert_date', 'disposition', 'disposition_time',
             'disposition_user_id', 'event_time']
        stats: The statistics you want returned.
        business_hours: None or the businesstime.BusinessTime representation
          of the business hours to use  for calculating time base
          statistics with business hours applied.  See `define_business_time()`.

    Returns:
        A dictionary where the key is a statistic found in VALID_ALERT_STATS 
        and the values are the calculated pandas.DataFrames for that statistic.
        Each statistical value is by month and disposition.
    """

    months = alerts.index.get_level_values('month').unique()

    dispositions = ACE_KILLCHAIN_DISPOSITIONS

    # extend with any new age dispositions found, in the alert data.
    # sort alphabetically before extending
    dispositions.extend(sorted([d for d in list(alerts['disposition'].unique()) if d not in ACE_KILLCHAIN_DISPOSITIONS]))
 
    stat_data_map = {'cycle_time_sum': {},
                     'cycle_time_mean': {},
                     'cycle_time_min': {},
                     'cycle_time_max': {},
                     'cycle_time_std': {},
                     'alert_count': {}
                    }

    logging.debug(f"Calculate statistics for {len(alerts)} alerts ")
    for dispo in dispositions:   
        for stat in stats:
            # define the space
            stat_data_map[stat][dispo] = {}
            
        for month in months:
            month_df = alerts.loc[month]
            
            # 1 dispo type during month means DataFrame selection gives a Series
            # alert_statistics_by_disposition expects a DataFrame
            if isinstance(month_df, pd.Series):
                month_df = pd.DataFrame([month_df])

            month_df = alert_statistics_by_disposition(month_df, business_hours)

            for stat in stats:
                try:
                    value = month_df.at[stat, dispo]
                except KeyError:
                    # dispo didn't happen during the given month
                    value = None
                    
                if isinstance(value, timedelta):
                    # convert to hours - XXX Make configurable?
                    value = value.total_seconds() / 60 / 60   

                stat_data_map[stat][dispo][month] = value

    for stat in stats:  
        stat_data_map[stat] = pd.DataFrame(data=stat_data_map[stat])
        stat_data_map[stat].fillna(0, inplace=True)
                
    if stat == 'alert_count':
        stat_data_map[stat] = stat_data_map[stat].astype(int)
        
    return stat_data_map

def organize_alerts_by_time_category(alerts: pd.DataFrame,
                                     business_hours: businesstime.BusinessTime
                                    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Organize alerts by 'week nights', 'weekends', and 'business hours'.

    Evaluate what time category each alert in alerts falls into. An alert
    was either created on week nights, weekends, or during business hours.
    Create a DataFrame for each time category and fill each DataFrame with the
    alerts that belong in it. This is accomplished by iterating over every alert
    and recording the alerts index number into a list for each time categorty. 
    Finally, the indexes are selected out of the original DataFrame to create
    the new DataFrames.

    Args:
        alerts: A pd.DataFrame of alerts
        business_hours: The businesstime.BusinessTime representation of the
          business hours to use. See `define_business_time()`

    Retruns:
        Tuple of pd.DataFrame objects in this order:
           weekend_df, nights_df, business_df
    """

    # unsetting to better intertuple
    alerts.reset_index(0, inplace=True)

    weekend_indexes = []
    bday_indexes = []
    night_indexes = []

    start_hour = business_hours._start_hour
    end_hour = business_hours._end_hour

    # track alert indexes
    i=0
    for row in alerts.itertuples():
        if row.insert_date.weekday() == 0: 
            # Monday
            if row.insert_date.time() < time(hour=start_hour, minute=0, second=0):
                # Before business hours -> the weekend
                weekend_indexes.append(i)
            elif row.insert_date.time() >= time(hour=end_hour, minute=0, second=0):
                # After business hours -> weeknight
                night_indexes.append(i)
            else: 
                # Buisness hours
                bday_indexes.append(i)

        elif row.insert_date.weekday() == 1:
            # Tuesday: either side of business hours -> week night
            if ( row.insert_date.time() < time(hour=start_hour, minute=0, second=0)
                 or row.insert_date.time() >= time(hour=end_hour, minute=0, second=0) ):
                night_indexes.append(i)
            else:
                # Buisness hours
                bday_indexes.append(i)

        elif row.insert_date.weekday() == 2:
            # Wednesday: either side of business hours -> week night
            if ( row.insert_date.time() < time(hour=start_hour, minute=0, second=0)
                 or row.insert_date.time() >= time(hour=end_hour, minute=0, second=0) ):
                night_indexes.append(i)
            else:
                # Buisness hours
                bday_indexes.append(i)

        elif row.insert_date.weekday() == 3:
            # Thursday: either side of business hours -> week night
            if ( row.insert_date.time() < time(hour=start_hour, minute=0, second=0)
                 or row.insert_date.time() >= time(hour=end_hour, minute=0, second=0) ):
                night_indexes.append(i)
            else:
                # Buisness hours
                bday_indexes.append(i)

        elif row.insert_date.weekday() == 4:
            # Friday
            if row.insert_date.time() < time(hour=start_hour, minute=0, second=0):
                # Before business hours -> weeknight
                night_indexes.append(i)
            elif row.insert_date.time() >= time(hour=end_hour, minute=0, second=0):
                # After business hours -> weekend
                weekend_indexes.append(i)
            else:
                # Buisness hours
                bday_indexes.append(i)

        elif row.insert_date.weekday() == 5: 
            # Saturday -> weekend bucket
            weekend_indexes.append(i)
        
        elif row.insert_date.weekday() == 6: 
            # Sunday -> weekend bucket
            weekend_indexes.append(i)

        else:
            logging.error("this should never happen")

        # next index
        i+=1

    weekend_df = alerts[alerts.index.isin(weekend_indexes)]
    bday_df = alerts[alerts.index.isin(bday_indexes)]
    nights_df = alerts[alerts.index.isin(night_indexes)]

    if((len(weekend_df) + len(nights_df) + len(bday_df)) != len(alerts) ):
        logging.critical("incorrect alert count.")

    return weekend_df, nights_df, bday_df

def generate_hours_of_operation_summary_table(alerts: pd.DataFrame,
                                              business_hours: businesstime.BusinessTime
                                             ) -> pd.DataFrame:
    """Cycle-time averages and alert quantities by operating hours and month.

    Summarize the overall cycle-time averages, the standard deviation in cycle times,
    and the alert quantities observed over the alerts passed when organized by month 
    and placed in the respective operating hour category. Categories are:
      business hours, weekends, weeknights

    Args:
        alerts: A pd.DataFrame of alerts
        business_hours: The businesstime.BusinessTime representation of the
          business hours to use. See `define_business_time()`

    Returns:
        A pd.DataFrame.	    
    """

    # operate on a copy
    alerts = alerts.copy()

    months = alerts.index.get_level_values('month').unique()

    weekend, nights, bday = organize_alerts_by_time_category(alerts, business_hours)
    weekend.set_index('month', inplace=True)
    nights.set_index('month', inplace=True)
    bday.set_index('month', inplace=True)

    if business_hours is None:
        logging.info(f"using default business time for business time dependent calculation.")
        # use defaults
        business_hours = define_business_time()

    start_hour = business_hours._start_hour
    end_hour = business_hours._end_hour

    business_day_cycle_time_averages = []
    weekend_cycle_time_averages = []
    nights_cycle_time_averages = []
    business_day_cycle_time_stdev = []
    weekend_cycle_time_stdev = []
    nights_cycle_time_stdev = []
    business_day_quantities = []
    weekend_quantities = []
    weeknight_quantities = []
    for month in months:
        try:
            bday_ct = bday.loc[month, 'disposition_time'] - bday.loc[month, 'insert_date']
        except KeyError:
            # month not in index, or only one alert
            bday_ct = pd.Series(data=pd.Timedelta(0))
        try:
            nights_ct = nights.loc[month, 'disposition_time'] - nights.loc[month, 'insert_date']
        except KeyError:
            # month not in index 
            nights_ct = pd.Series(data=pd.Timedelta(0))
        try:
            weekend_ct = weekend.loc[month, 'disposition_time'] - weekend.loc[month, 'insert_date']
        except KeyError:
            weekend_ct = pd.Series(data=pd.Timedelta(0))

        # handle case of single alert in a bucket for the month
        if isinstance(bday_ct, pd.Timedelta):
            bday_ct = pd.Series(data=bday_ct)
        if isinstance(nights_ct, pd.Timedelta):
            nights_ct = pd.Series(data=nights_ct)
        if isinstance(weekend_ct, pd.Timedelta):
            weekend_ct = pd.Series(data=weekend_ct)

        business_day_cycle_time_averages.append((bday_ct.mean().total_seconds() / 60) / 60)
        nights_cycle_time_averages.append((nights_ct.mean().total_seconds() / 60) / 60)
        weekend_cycle_time_averages.append((weekend_ct.mean().total_seconds() / 60) / 60)

        business_day_cycle_time_stdev.append((bday_ct.std().total_seconds() / 60) / 60)
        weekend_cycle_time_stdev.append((nights_ct.std().total_seconds() / 60) / 60)
        nights_cycle_time_stdev.append((weekend_ct.std().total_seconds() / 60) / 60)

        business_day_quantities.append(len(bday_ct))
        weeknight_quantities.append(len(nights_ct))
        weekend_quantities.append(len(weekend_ct))

    data = {
             ('Cycle-Time Averages', 'Business'): business_day_cycle_time_averages,
             ('Cycle-Time Averages', 'Nights'): nights_cycle_time_averages,
             ('Cycle-Time Averages', 'Weekend'): weekend_cycle_time_averages,
             ('Cycle-Time Std. Dev.', 'Business'): business_day_cycle_time_stdev,
             ('Cycle-Time Std. Dev.', 'Nights'): nights_cycle_time_stdev,
             ('Cycle-Time Std. Dev.', 'Weekend'): weekend_cycle_time_stdev,
             ('Quantities', 'Bus Hrs'): business_day_quantities,
             ('Quantities', 'Nights'): weeknight_quantities,
             ('Quantities', 'Weekend'): weekend_quantities
            }

    hop_df = pd.DataFrame(data, index=months)
    hop_df.fillna(0, inplace=True)
    hop_df.name = "Hours of Operation"
    return hop_df

def generate_overall_summary_table(alerts: pd.DataFrame,
                                   business_hours: businesstime.BusinessTime
                                  ) -> pd.DataFrame:
    """Generate an overall statistical summary for alerts by month.

    Organize alerts by month and then summarize the business hour and real hour
    cycle-time averages, the standard deviation in the business hour and real 
    hour cycle-times, and the alert quantities observed over the alerts passed.

    Args:
        alerts: A pd.DataFrame of alerts
        business_hours: The businesstime.BusinessTime representation of the
          business hours to use. See `define_business_time()`

    Returns:
        A pd.DataFrame.	    
    """

    months = alerts.index.get_level_values('month').unique()

    quantities = []
    bh_cycletime = []
    bh_std = []
    real_cycletime = []
    real_std = []
    for month in months:
        bh_alert_ct = get_business_hour_cycle_time(alerts.loc[[month],['disposition_time', 'insert_date']],
                                                   business_hours)
        alert_ct = alerts.loc[month, 'disposition_time'] - alerts.loc[month, 'insert_date']
        quantities.append(len(alerts.loc[month]))

        if isinstance(bh_alert_ct, pd.Timedelta):
            bh_alert_ct = pd.Series(data=bh_alert_ct)
        if isinstance(alert_ct, pd.Timedelta):
            alert_ct = pd.Series(data=alert_ct)
        bh_cycletime.append(((bh_alert_ct.mean()).total_seconds() / 60 ) / 60 )
        bh_std.append(((bh_alert_ct.std()).total_seconds() / 60) / 60 )
        real_cycletime.append(((alert_ct.mean()).total_seconds() / 60 ) / 60 )
        real_std.append(((alert_ct.std()).total_seconds() / 60) / 60 )

    data = {
             'Business hour Cycle Time Avg.': bh_cycletime,
             'Business hour Cycle Time Std.': bh_std,
             'Cycle time Avg.': real_cycletime,
             'Cyclet time std.': real_std,
             'Quantity': quantities
           }

    result = pd.DataFrame(data, index=months)
    result.fillna(0, inplace=True)
    result.name = "Overall Operating Alert Summary"
    return result


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

__version__ = '0.1.0'
description = 'A lib for measuring ACE based IDR operations.'

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Load requirements
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    requirements = [line.strip() for line in f.readlines()]

setup(
    name='ace-metrics',

    version=__version__,

    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/seanmcfeely/ace-metrics',
    author='Sean McFeely',
    author_email='mcfeelynaes@gmail.com',
    license='GNU General Public License v3.0',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
    ],
    python_requires='>=3.6',
    keywords='Information Security,ACE,ACE Ecosystem',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    scripts=['ace-metrics']
)




"""Functions that are helpful to other metric funcitons."""

import io
import os
import logging
import pymysql
import tarfile
import configparser

import pandas as pd
from datetime import timedelta, datetime, time
from dateutil.relativedelta import relativedelta

from typing import Mapping, List, Tuple

from .alerts import FRIENDLY_STAT_NAME_MAP

CompanyID = int
CompanyName = str
CompanyMap = Mapping[CompanyID, CompanyName]

def generate_html_plot(data_table: pd.DataFrame,
                       kind="line",
                       legend="top_left",
                       toolbar_location="above",
                       xlabel="Month",
                       figsize=(1000,600),
                       title=None,
                       ylabel=None) -> str:
    """Convert a datatable into an HTML Bokeh plot.

    This is very customized with defaults.

    Plot kinds from pandas_bokeh:
    https://github.com/PatrikHlobil/Pandas-Bokeh/blob/1a8309ca7b5cbee527cf951668e4f8a1250cecb3/pandas_bokeh/plot.py#L167

    Args:
        data_table: A pandas DataFrame

    Returns:
        An HTML element representing the plot,
        as a string.
    """
    import math
    import pandas_bokeh

    if ylabel is None:
        ylabel="Hours"

    if title is None:
        try:
            title = data_table.name
        except AttributeError:
            title = ""    

    p = data_table.plot_bokeh(kind=kind,
                              show_figure=False,
                              legend=legend,
                              toolbar_location=toolbar_location,
                              figsize=figsize,
                              title=title,
                              zooming=False,
                              xlabel=xlabel,
                              ylabel=ylabel)

    # override legend defaults
    p.legend.background_fill_alpha = 0
    p.legend.border_line_alpha = 0
    p.xaxis.major_label_orientation = math.pi/4
    if ylabel is not None and 'alert' in title.lower():
        p.yaxis.axis_label = "Number of Alerts"
    return pandas_bokeh.embedded_html(p)

def get_companies(con: pymysql.connections.Connection) -> CompanyMap:
    """Query the database for all companies.

    Args:
        con: a pymysql database connectable

    Returns:
        A dict of companies like so:
          {company_id: 'company_name'}, ..
    """

    companies = {}
    cursor = con.cursor()
    cursor.execute("select * from company")
    for c_id,c_name in cursor.fetchall():
        companies[c_id] = c_name
    return companies

def apply_company_selection_to_query(query: str, company_ids: list, selected_companies: list) -> str:
    """Update a metric SQL query to select where companies.

    Args:
        query: An ACE DB query structered for reduction by company ID.
          Such a query should have two "{}" back to back, like: {}{}
        company_ids: list of all valid company IDs
        selected_companies: A list of companies to select alerts for, by name.
          If the list is empty, all alerts are selected.

    Returns:
        An updated SQL query string.

    """
    return query.format(' AND ' if company_ids else '', '( ' + ' OR '.join(['company.name=%s' for company in selected_companies]) +') ' if company_ids else '')

def sanitize_table_name(table_name=None, keep_friendly=False) -> str:
    """Sanitize a table name.

    Try to make a table name safe to be a file name.

    Args:
        table_name: The table name to sanitize.
        keep_friendly: Do not map detailed friendly statistic names back
          to their less descriptive key names.

    Returns:
       A santized name string.
    """

    if table_name is None:
        return f"No name - {datetime.now().timestamp()}"

    safe_name = table_name.strip()

    if not keep_friendly:
        # map the friendly names back to key name
        for stat_key,stat_name in FRIENDLY_STAT_NAME_MAP.items():
            if stat_name in safe_name:
                safe_name = safe_name.replace(stat_name, stat_key)

    _invalid_chars = ["\\", "*", "?", ":", "/", "[", "]"]
    for invalid_char in _invalid_chars:
        safe_name = safe_name.replace(invalid_char, '-')

    return safe_name

def dataframes_to_archive_bytes_of_json_files(tables: List[pd.DataFrame]) -> bytes:
    """Create byte archive of tables as json files.

    Convert each table to its json file bytestring. Put each bytestring
    into a tar archive and return the tar.gz bytes of that archive.
    Write to a file or send wherever.

    Args:
        tables: A list of pd.DataFrames

    Returns:
        The bytestring of a tar.gz archive containing the
        tables as json files.
    """

    buf = io.BytesIO()
    tar = tarfile.open(mode="w:gz", fileobj=buf)
    for table in tables:
        if not table.name:
            safe_table_name = sanitize_table_name()
        else:
            safe_table_name = sanitize_table_name(table.name, keep_friendly=True)

        table_buf = io.BytesIO()
        table_bytes = table.to_json().encode('utf-8')
        table_info = tarfile.TarInfo(name=f"{safe_table_name}.json")
        table_info.size = len(table_bytes)
        table_buf.write(table_bytes)
        table_buf.seek(0)
        tar.addfile(table_info, table_buf)
        table_buf.close()

    tar.close()
    buf.seek(0)
    filebytes = buf.read()
    buf.close()
    return filebytes

def dataframes_to_xlsx_bytes(tables: List[pd.DataFrame]) -> bytes:
    """Export dataframes to xlsx bytes.

    Write the bytes to a file or send them wherever.

    Args:
        tables: A list of pd.DataFrames

    Returns:
        The bytestring representation of the xlsx file.
    """

    tab_names = []
    tab_name_map = {}
    table_tab_map = {}
    # sanitize and make tab name map
    for table in tables:
        if table.name:
            table_name = table.name
        else:
            logging.warning("metric table has no name.")
            table_name = sanitize_table_name()
        clean_table_name = sanitize_table_name(table.name)

        # do additional table name cleanup for excel
        # try to clean up alert_type names
        name_parts = clean_table_name.split(' - ')
        if name_parts:
            _tmp_name = ""
            for part in name_parts[:-1]:
                _tmp_name += f"{part[0].upper()}-"
            clean_table_name = f"{_tmp_name}{name_parts[-1]}"

        # openpyxl guidance to keep names to 31 chars or less
        if len(clean_table_name) > 31:
            clean_table_name = clean_table_name[:31]

        if clean_table_name in tab_names:
            logging.warning(f"name collision for {clean_table_name}")
            # 30 char collision name
            clean_table_name = f"Collision - {datetime.now().timestamp()}"

        tab_names.append(clean_table_name)

        logging.info(f"changed table name from '{table_name}' to '{clean_table_name}'")
        # will add this helpful info to the excel sheet
        tab_name_map[clean_table_name] = table_name
        table_tab_map[clean_table_name] = table

    xlsx_bytes = io.BytesIO()
    writer = pd.ExcelWriter(xlsx_bytes)
    # write the tab name map first
    tab_name_map_df = pd.DataFrame.from_dict(tab_name_map,
                                            orient='index',
                                            columns=['ACE Data Table Name'])
    tab_name_map_df.index.names = ['Tab Name']
    tab_name_map_df.to_excel(writer, "Tab Name Map")

    # write the tables to excel tabs
    for name, table in table_tab_map.items():
        try:
            table.to_excel(writer, name)
        except Exception as e:
            logging.error(f"failed to write table: {e}")

    writer.close()
    xlsx_bytes.seek(0)
    filebytes = xlsx_bytes.read()
    xlsx_bytes.close()

    return filebytes

def connect_to_database(config: configparser.SectionProxy) -> pymysql.connections.Connection:
    """Connect to a configured ACE DB.

    Args:
        config: A configparser section that defines
          the database connection.

    Returns:
        pymysql.connections.Connection to an ACE DB
    """
    from getpass import getpass

    ssl_settings = None
    if os.path.exists(config.get('ssl_ca_path')):
        ssl_settings = {'ca': config['ssl_ca_path']}

    password = config.get('pass')
    if not password:
        password = getpass(f"Enter password for {config['host']}: ")

    db = pymysql.connect(host=config['host'], user=config['user'], password=password, database=config['database'], ssl=ssl_settings)
    return db


def get_month_keys_between_two_dates(start_date: datetime, end_date: datetime) -> list:
    """Get unique %Y%m (Months) between dates."""

    months = []
    while start_date.year <= end_date.year:
        while start_date.month <= end_date.month:
            months.append(datetime.strftime(start_date, '%Y%m'))
            start_date += relativedelta(months=1)
            break

    return months



import os
import sys
import json
import logging
import configparser
import datetime
import pymysql
import argparse
import argcomplete

import pandas as pd

from tabulate import tabulate

from .alerts import ( VALID_ALERT_STATS, 
                      FRIENDLY_STAT_NAME_MAP,
                      ALERTS_BY_MONTH_DB_QUERY,
                      statistics_by_month_by_dispo )


STDOUT_FORMATS = ['json', 'csv', 'ascii_table', 'print']

FILEOUT_FORMATS = ['json', 'xlsx']#, 'csv', 'sqlite']

def stdout_like(df: pd.DataFrame, format='print'):
    if format not in STDOUT_FORMATS:
        logging.warning(f"{format} is not a supported output format for the cli")
        return False

    table_name = ""
    try:
        table_name = f"{df.name}:"
    except AttributeError:
        # table has no name
        pass

    if format == 'json':
        print(df.to_json())
        return

    if format == 'csv':
        print(df.to_csv())
        return

    if format == 'ascii_table':
        print()
        print(table_name)
        print(tabulate(df, headers='keys', tablefmt='simple'))
        print()
        return
    
    print()
    print(table_name)
    print(df)
    print()
    return

def build_metric_user_parser(user_parser: argparse.ArgumentParser) -> None:
    """Given an argparse subparser, build a metric user parser.

    Args:
        user_parser: An argparse.ArgumentParser.

    Returns: None
    """

    user_parser.add_argument('-l', '--list-users', action='store_true',
                              help='List all users')
    user_parser.add_argument('-u', '--user', action='append', dest='users', default=[],
                             help='A list of users to generate statistics for. Default: All users.')
    user_parser.add_argument('-tac', '--total-alert-count-breakdown',
                             action='store_true', help="Total count of Alerts worked by analyst by month.")

    for stat in VALID_ALERT_STATS:
        user_parser.add_argument(f'--{stat}', action='store_true', dest=f"user_stat_{stat}",
                                 help=FRIENDLY_STAT_NAME_MAP[stat])
    user_parser.add_argument('--all-stats', dest='all_user_stats', action='store_true',
                             help="Return all of the available statistics.")

def build_metric_alert_type_parser(alert_type_parser: argparse.ArgumentParser) -> None:
    """Given an argparse subparser, build a metric alert type parser.

    Args:
        alert_type_parser: An argparse.ArgumentParser.

    Returns: None
    """

    alert_type_parser.add_argument('-l', '--list-alert-types', action='store_true',
                              help='List the types of alerts')
    alert_type_parser.add_argument('-t', '--type', action='append', dest='types', default=[],
                             help='A list of alert_types to generate statistics for. Default: All alert types.')
    alert_type_parser.add_argument('-c', '--overall-count-breakdown', action='store_true',
                             help='An overall breakdown of alert counts by alert type.')

    for stat in VALID_ALERT_STATS:
        alert_type_parser.add_argument(f'--{stat}', action='store_true', dest=f"alert_type_stat_{stat}",
                                       help=FRIENDLY_STAT_NAME_MAP[stat])
    alert_type_parser.add_argument('--all-stats', dest='all_at_stats', action='store_true',
                                   help="Return all of the available statistics.")

def build_metric_alert_parser(alert_parser: argparse.ArgumentParser) -> None:
    """Given an argparse subparser, build a metric alert parser.

    Build an alert parser that defines how to interface with the
    ACE metrics library for ACE alert data.
    
    Args:
        alert_parser: An argparse.ArgumentParser.

    Returns: None
    """

    alert_parser.add_argument('-hop', '--hours-of-operation', action='store_true',
                              help='Generate "Hours of Operation" summary by month')
    alert_parser.add_argument('-avg-ct-sum', '--average-alert-cycletime-summary', action='store_true',
                              help="Overall summary of alert cycle times by month" )

    for stat in VALID_ALERT_STATS:
        alert_parser.add_argument(f'--{stat}', action='store_true', dest=f"alert_stat_{stat}",
                                  help=FRIENDLY_STAT_NAME_MAP[stat])
    alert_parser.add_argument('--all-stats', dest='all_alert_stats', action='store_true',
                              help="Return all of the available statistics.")

    alert_subparsers = alert_parser.add_subparsers(dest='alert_metric_target')

    user_parser = alert_subparsers.add_parser("users", help="user based alert metrics")
    build_metric_user_parser(user_parser)

    alert_type_parser = alert_subparsers.add_parser("types", help="alert metrics by alert types")
    build_metric_alert_type_parser(alert_type_parser)

def build_metric_event_parser(event_parser: argparse.ArgumentParser) -> None:
    """Given an argparse subparser, build a metric event parser.

    Build an event parser that defines how to interface with the
    ACE metrics library for ACE event data.
    
    Args:
        event_parser: An argparse.ArgumentParser.

    Returns: None
    """ 

    event_parser.add_argument('-i', '--incidents', action='store_true',
                              help='Create incident table from event table')
    event_parser.add_argument('-io', '--incidents-only', action='store_true',
                              help='Only return incidents table (skip event table)')
    event_parser.add_argument('-ce', '--count-emails', action='store_true',
                              help='Count emails, in each event, per company.')

def build_metric_parser(parser: argparse.ArgumentParser) -> None:
    """Build the ACE metric parser.
    
    Args:
        parser: An argparse.ArgumentParser.

    Returns: None
    """

    # Default date range will be the last 7 days.
    default_start_datetime = (datetime.datetime.today() - datetime.timedelta(days=7)).strftime('%Y-%m-%d %H:%M:%S')
    default_end_datetime = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')

    parser.add_argument('-so', '--stdout-format', default='print', action='store', choices=STDOUT_FORMATS,
                        help="desired standard output format. ~~~ NOTE: 'print' (the default) will also summarize large tables. Use 'ascii_table' to avoide that.")
    parser.add_argument('-fo', '--fileout-format', default=None, action='store', choices=FILEOUT_FORMATS,
                        help="desired file output format. Default is xls.")
    parser.add_argument('-f', '--filename', action='store', default=None, help="The name of a file to write results to.")
    parser.add_argument('-c', '--company', action='append', dest='companies', default=[],
                        help="A list of company names to gather metrics for. Default is all defined companies.")
    parser.add_argument('-bh', '--business-hours', action='store', default=False,
                        help="Use business hours for all time based stats. Set like start_hour,end_hour,time_zone. Example: 6,18,US/Eastern")
    parser.add_argument('-s', '--start_datetime', action='store', default=default_start_datetime,
                        help="The start datetime data is in scope. Format: YYYY-MM-DD HH:MM:SS. Default: 7 days ago.")
    parser.add_argument('-e', '--end_datetime', action='store', default=default_end_datetime,
                         help="The end datetime data is in scope. Format: YYYY-MM-DD HH:MM:SS. Default: now.")

    metrics_subparsers = parser.add_subparsers(dest='metric_target')
 
    alert_parser = metrics_subparsers.add_parser("alerts", help="alert based metrics")
    build_metric_alert_parser(alert_parser)

    event_parser = metrics_subparsers.add_parser("events", help="event based metrics. With no arguments will return all events")
    build_metric_event_parser(event_parser)


def create_histogram_string(data: dict) -> str:
    """A convenience function that creates a graph in the form of a string.

    Args:
        data: A dictionary, where the values are integers representing a count of the keys.

    Returns:
        A graph in string form, pre-formatted for raw printing.
    """

    assert isinstance(data, dict)
    for key in data.keys():
        assert isinstance(data[key], int)

    total_results = sum([value for value in data.values()])
    txt = ""

    # order keys for printing in order (purly ascetics)
    ordered_keys = sorted(data, key=lambda k: data[k])
    results = []

    # longest_key used to calculate how many white spaces should be printed
    # to make the graph columns line up with each other
    longest_key = 0
    for key in ordered_keys:
        value = data[key]
        longest_key = len(key) if len(key) > longest_key else longest_key
        # IMPOSING LIMITATION: truncating keys to 95 chars, keeping longest key 5 chars longer
        longest_key = 100 if longest_key > 100 else longest_key
        percent = value / total_results * 100
        results.append((key[:95], value, percent, u"\u25A0"*(int(percent/2))))

    # two for loops are ugly, but allowed us to count the longest_key -
    # so we loop through again to print the text
    for r in results:
        txt += "%s%s: %5s - %5s%% %s\n" % (int(longest_key - len(r[0]))*' ', r[0] , r[1],
                                            str(r[2])[:4], u"\u25A0"*(int(r[2]/2)))
    return txt

def execute_metric_arguments(db: pymysql.connections.Connection, args):
    """Logic for executing metric CLI arguments.
    """

    from .alerts import ( get_alerts_between_dates,
                                 VALID_ALERT_STATS, 
                                 FRIENDLY_STAT_NAME_MAP,
                                 statistics_by_month_by_dispo,
                                 generate_hours_of_operation_summary_table,
                                 generate_overall_summary_table,
                                 define_business_time
                                )
    from .alerts.users import ( get_all_users,
                                generate_user_alert_stats,
                                alert_quantities_by_user_by_month )
    from .alerts.alert_types import ( unique_alert_types_between_dates,
                                             count_quantites_by_alert_type,
                                             get_alerts_between_dates_by_type,
                                             generate_alert_type_stats,
                                             all_alert_types
                                            )

    from .events import ( get_events_between_dates,
                                 get_incidents_from_events,
                                 add_email_alert_counts_per_event
                                )

    from .helpers import ( get_companies,
                                  dataframes_to_archive_bytes_of_json_files,
                                  dataframes_to_xlsx_bytes
                                )

    # store tables for printing or writing
    tables = []

    if args.companies:
        valid_companies = []
        valid_companies = list(get_companies(db).values())

        for company in args.companies:
            if company not in valid_companies:
                logging.warning(f"{company} is not a valid company: {valid_companies}")
                args.companies.remove(company)
        if not args.companies:
            sys.exit(0)

    if args.business_hours:
        bh_settings = args.business_hours.split(',')
        if not bh_settings:
            logging.error(f"invalid business hour format: {args.business_hours}")
            sys.exit(1)
        start_hour = int(bh_settings[0])
        end_hour = int(bh_settings[1])
        time_zone = bh_settings[2]
        # NOTE: holidays remain default
        args.business_hours = define_business_time(start_hour, end_hour, time_zone)

    if args.metric_target == 'events':
        start_date = datetime.datetime.strptime(args.start_datetime, '%Y-%m-%d %H:%M:%S')
        end_date = datetime.datetime.strptime(args.end_datetime, '%Y-%m-%d %H:%M:%S')
        events = get_events_between_dates(start_date, end_date, db, selected_companies=args.companies)
        if args.count_emails:
            add_email_alert_counts_per_event(events, db)

        if args.incidents or args.incidents_only:
            incidents = get_incidents_from_events(events)
            tables.append(incidents)

        if not args.incidents_only:
            events.drop(columns=['id'], inplace=True)
            tables.append(events)

    if args.metric_target == 'alerts':
        # The intention is to only get data that's needed while
        # allowing the user to express the widest range of possible options.

        users = {}
        user_ids = []
        if args.alert_metric_target == 'users':
            users = get_all_users(db)

            if args.list_users:
                analysts = pd.DataFrame.from_dict(users, orient='index')
                analysts.name = "Users"
                tables.append(analysts)

            if args.users:
                # only use specified users
                for username in args.users:
                    user_ids.extend([user_id for user_id, user in users.items() if username == user['username']])
            else:
                user_ids = [user_id for user_id in users.keys()]

            if args.total_alert_count_breakdown:
                start_date = datetime.datetime.strptime(args.start_datetime, '%Y-%m-%d %H:%M:%S')
                end_date = datetime.datetime.strptime(args.end_datetime, '%Y-%m-%d %H:%M:%S')
                user_dispositions_per_month = alert_quantities_by_user_by_month(start_date, end_date, db)
                if args.users:
                    tables.append(user_dispositions_per_month[args.users])
                else:
                    tables.append(user_dispositions_per_month)

        if args.alert_metric_target == 'types':
            if args.overall_count_breakdown:
                start_date = datetime.datetime.strptime(args.start_datetime, '%Y-%m-%d %H:%M:%S')
                end_date = datetime.datetime.strptime(args.end_datetime, '%Y-%m-%d %H:%M:%S')
                at_counts = count_quantites_by_alert_type(start_date, end_date, db)
                tables.append(at_counts)

            # does the user want specific alert_type statistics?
            at_arg_stats = [sarg[len('alert_type_stat_'):] for sarg, value in vars(args).items() if sarg.startswith('alert_type_stat_') and value is True]

            if args.list_alert_types or args.all_at_stats or at_arg_stats:
                start_date = datetime.datetime.strptime(args.start_datetime, '%Y-%m-%d %H:%M:%S')
                end_date = datetime.datetime.strptime(args.end_datetime, '%Y-%m-%d %H:%M:%S')
                alert_types = unique_alert_types_between_dates(start_date, end_date, db)

                if args.list_alert_types:
                    print(f"Alert Types between '{args.start_datetime}' and '{args.end_datetime}':")
                    [print(f"\t{at}") for at in alert_types]
                    print()
                    print("All Alert Types:")
                    [print(f"\t{at}") for at in all_alert_types(db)]

                if args.types:
                    # only use specified alert types
                    alert_types = [at for at in alert_types if at in args.types]

                if args.all_at_stats or at_arg_stats:
                    start_date = datetime.datetime.strptime(args.start_datetime, '%Y-%m-%d %H:%M:%S')
                    end_date = datetime.datetime.strptime(args.end_datetime, '%Y-%m-%d %H:%M:%S')
                    alert_type_map = get_alerts_between_dates_by_type(start_date, end_date, db, selected_companies=args.companies)

                    alert_type_stat_map = generate_alert_type_stats(alert_type_map, business_hours=args.business_hours)

                    for alert_type in alert_types:
                        if args.all_at_stats:
                            for stat in VALID_ALERT_STATS:
                                tables.append(alert_type_stat_map[alert_type][stat])
                        else:
                            for stat in at_arg_stats:
                                tables.append(alert_type_stat_map[alert_type][stat])

        # does the user want specific alert statistics?
        alert_arg_stats = [sarg[len('alert_stat_'):] for sarg, value in vars(args).items() if sarg.startswith('alert_stat_') and value is True]

        # only get alert data if we really needs it
        if args.alert_metric_target == 'users' or alert_arg_stats \
          or args.all_alert_stats or args.hours_of_operation or args.average_alert_cycletime_summary:
            start_date = datetime.datetime.strptime(args.start_datetime, '%Y-%m-%d %H:%M:%S')
            end_date = datetime.datetime.strptime(args.end_datetime, '%Y-%m-%d %H:%M:%S')
            alerts = get_alerts_between_dates(start_date, end_date, db, selected_companies=args.companies)

        if args.alert_metric_target == 'users':
            all_user_stat_map = generate_user_alert_stats(alerts, users, business_hours=args.business_hours)

            for user_id in user_ids:
                if args.all_user_stats:
                    for stat in VALID_ALERT_STATS:
                        tables.append(all_user_stat_map[user_id][stat])
                else:
                    user_arg_stats = [sarg[len('user_stat_'):] for sarg, value in vars(args).items() if sarg.startswith('user_stat_') and value is True]
                    for stat in user_arg_stats:
                        tables.append(all_user_stat_map[user_id][stat])

        if alert_arg_stats or args.all_alert_stats:
            alert_stat_map = statistics_by_month_by_dispo(alerts, business_hours=args.business_hours)

            if args.all_alert_stats:
                for stat in VALID_ALERT_STATS:
                    alert_stat_map[stat].name = FRIENDLY_STAT_NAME_MAP[stat]
                    tables.append(alert_stat_map[stat])
            else:
                for stat in alert_arg_stats:
                    alert_stat_map[stat].name = FRIENDLY_STAT_NAME_MAP[stat]
                    tables.append(alert_stat_map[stat])

        if args.hours_of_operation:
            if not args.business_hours:
                # business hours are required
                args.business_hours = define_business_time()
            hop_df = generate_hours_of_operation_summary_table(alerts.copy(), args.business_hours)
            if args.fileout_format:
                tables.append(hop_df)
            else:
                stdout_like(hop_df, format=args.stdout_format)

        if args.average_alert_cycletime_summary:
            if not args.business_hours:
                # business hours are required
                args.business_hours = define_business_time()
            sla_df = generate_overall_summary_table(alerts.copy(), args.business_hours)
            if args.fileout_format:
                tables.append(sla_df)
            else:
                stdout_like(sla_df, format=args.stdout_format)

    # output
    if args.fileout_format and tables:
        time_stamp = str(datetime.datetime.now().timestamp())
        time_stamp = time_stamp[:time_stamp.rfind('.')]
        filename = f"ACE_metrics_{time_stamp}"

        if args.fileout_format == 'xls':
            if args.filename:
                filename = args.filename
            else:
                filename += ".xlsx"
            with open(filename, 'wb') as fp:
                fp.write(dataframes_to_xlsx_bytes(tables))
            if os.path.exists(filename):
                print(f" + wrote {filename}")
        if args.fileout_format == 'json':
            if args.filename:
                filename = args.filename
            else:
                filename += ".tar.gz"
            with open(filename, 'wb') as fp:
                fp.write(dataframes_to_archive_bytes_of_json_files(tables))
            if os.path.exists(filename):
                print(f" + wrote {filename}")
    else:
        for table in tables:
            stdout_like(table, format=args.stdout_format)

    db.close()

def cli_metrics():
    """Main entry point for metrics on the CLI.
    """

    from .helpers import connect_to_database
    from .config import load_config, DEFAULT_CONFIG_PATHS

    parser = argparse.ArgumentParser(description="CLI Interface to ACE Metrics")
    parser.add_argument('-d', '--debug', action='store_true', help="Turn on debug logging.")
    parser.add_argument('-cp', '--config-path', action='store', help="Path to config defining ACE db connection.")

    build_metric_parser(parser)

    argcomplete.autocomplete(parser)
    args = parser.parse_args()

    config_paths = DEFAULT_CONFIG_PATHS
    if args.config_path:
        if not os.path.exists(args.config_path):
            logging.error(f"{args.config_path} does not exist.")
            return False
        config_paths.append(args.config_path) 

    config = load_config(config_paths)
    if not config:
        logging.error("You must define and supply a configuration. See `ace-metrics -h | grep config`")
        return False

    database_config = config['database_default']

    db = connect_to_database(database_config)

    return execute_metric_arguments(db, args)



"""Constants for event metrics."""

# The core ACE specific killchain dispositions that
# mean an event should be considered an incident.
# If you add more, add them to the end of the list or
# extend this list where ever you need to.
INCIDENT_DISPOSITIONS = [ 'EXPLOITATION',
                          'INSTALLATION',
                          'COMMAND_AND_CONTROL',
                          'EXFIL',
                          'DAMAGE'
                        ]

# The core ACE specific killchain dispositions that
# mean an event should be considered an event.
EVENT_DISPOSITIONS = ['DELIVERY'] + INCIDENT_DISPOSITIONS

# Database query for getting events between two dates
#  and pulling relevant details from the various table mappings
# Allows for reduction by list of company ids
EVENT_DB_QUERY = """SELECT 
                        events.id, 
                        events.creation_date as 'Date', events.name as 'Event', 
                        GROUP_CONCAT(DISTINCT malware.name SEPARATOR ', ') as 'Malware', 
                        GROUP_CONCAT(DISTINCT IFNULL(malware_threat_mapping.type, 'UNKNOWN') SEPARATOR ', ') 
                            as 'Threat', GROUP_CONCAT(DISTINCT alerts.disposition SEPARATOR ', ') as 'Disposition', 
                        events.vector as 'Delivery Vector', 
                        events.prevention_tool as 'Prevention', 
                        GROUP_CONCAT(DISTINCT company.name SEPARATOR ', ') as 'Company', 
                        count(DISTINCT event_mapping.alert_id) as '# Alerts' 
                    FROM events 
                        JOIN event_mapping 
                            ON events.id=event_mapping.event_id 
                        JOIN malware_mapping 
                            ON events.id=malware_mapping.event_id 
                        JOIN malware 
                            ON malware.id=malware_mapping.malware_id 
                        JOIN company_mapping 
                            ON events.id=company_mapping.event_id 
                        JOIN company 
                            ON company.id=company_mapping.company_id 
                        LEFT JOIN malware_threat_mapping 
                            ON malware.id=malware_threat_mapping.malware_id 
                        JOIN alerts 
                            ON alerts.id=event_mapping.alert_id 
                    WHERE 
                        events.status='CLOSED' AND events.creation_date 
                        BETWEEN %s AND %s {}{}
                    GROUP BY events.name, events.creation_date, event_mapping.event_id 
                    ORDER BY events.creation_date
                """

# Database query for getting events between two dates
# and organized by a time period (week/month) for the purpose
# of counting. Allows for reduction by list of company ids.
EVENT_COUNT_TIME_DB_QUERY = """SELECT
                        events.id, DATE_FORMAT(events.creation_date, '{<[TIME_FORMAT]>}') as '{<[TIME_KEY]>}',
                        events.creation_date as 'Date', events.name as 'Event',
                        GROUP_CONCAT(DISTINCT alerts.disposition SEPARATOR ', ') as 'Disposition',
                        GROUP_CONCAT(DISTINCT company.name SEPARATOR ', ') as 'Company',
                        count(DISTINCT event_mapping.alert_id) as '# Alerts'
                    FROM events
                        JOIN event_mapping
                            ON events.id=event_mapping.event_id
                        JOIN company_mapping
                            ON events.id=company_mapping.event_id
                        JOIN company
                            ON company.id=company_mapping.company_id
                        JOIN alerts
                            ON alerts.id=event_mapping.alert_id
                    WHERE
                        events.status='CLOSED' AND events.creation_date
                        BETWEEN %s AND %s {}{}
                    GROUP BY events.name, events.creation_date, event_mapping.event_id
                    ORDER BY events.creation_date
                """




"""Metric functions for ACE Events data.

Every function in this file should expliclty work with
ACE database tables related to events for the purpose 
of generating metrics.

"""

import os
import logging

import pymysql
import pandas as pd

from datetime import datetime

from .constants import ( INCIDENT_DISPOSITIONS,
                         EVENT_DISPOSITIONS,
                         EVENT_DB_QUERY,
                         EVENT_COUNT_TIME_DB_QUERY
                       )

def add_email_alert_counts_per_event(events: pd.DataFrame,
                                     con: pymysql.connections.Connection,
                                    ) -> pd.DataFrame:
    """Count the number of emails in each event.

    Iterate over every event and count the number of alerts, in each event,
    with a message_id associated to the event. The passed events pd.DataFrame
    is edited in place to include an additional column that indicates the count
    of emails in the event. If there is more than one company, in the event,
    comma seperate the email counts by company.

    NOTE: This function should be improved by only counting unique message_ids.

    Args:
        events: A pd.DataFrame of events
        con: a pymysql database connectable

    Returns:
        None
    """

    # given event id and company name, get alert count per company.
    alert_count_per_company_query = """SELECT 
        COUNT(DISTINCT event_mapping.alert_id) as 'alert_count' 
        FROM event_mapping 
        JOIN alerts 
            ON alerts.id=event_mapping.alert_id 
        LEFT JOIN company 
            ON company.id=alerts.company_id 
        WHERE 
        event_mapping.event_id=%s AND company.name=%s"""

    # given event id and company name, get count of emails based on 
    # alerts with a message_id observables.
    # NOTE: I don't remember why `a.alert_type!='o365'` is specified
    #   should probably be removed.
    email_count_per_company_query = """SELECT
                a.id, a.alert_type, o.value 
            FROM
                observables o
                JOIN observable_mapping om ON om.observable_id = o.id
                JOIN alerts a ON a.id = om.alert_id
                JOIN event_mapping em ON em.alert_id=a.id
                JOIN company c ON c.id = a.company_id
            WHERE
                o.type = 'message_id'
                AND em.event_id=%s
                AND a.alert_type!='o365'
                AND c.name=%s"""

    email_counts = []
    for event in events.itertuples():
        if ',' in event.Company:
            companies = event.Company.split(', ')
            new_alert_count = email_count = ""
            for company in companies:
                params = [event.id, company]
                company_alerts = pd.read_sql_query(alert_count_per_company_query, con, params=params)
                email_alerts = pd.read_sql_query(email_count_per_company_query, con, params=params)
                new_alert_count += str(int(company_alerts.alert_count.values))+","

                # all mailbox alerts will be unique phish
                mailbox_phish = email_alerts.loc[email_alerts.alert_type=='mailbox']
                mailbox_phish_count = len(mailbox_phish)
                unique_phish = list(set(mailbox_phish.value.values))

                # remove mailbox alerts and leave any other alerts with a message_id observable
                email_alerts = email_alerts[email_alerts.alert_type!='mailbox']
                for alert in email_alerts.itertuples():
                    if alert.value not in unique_phish:
                        unique_phish.append(alert.value)
                        mailbox_phish_count += 1
                email_count += str(mailbox_phish_count)+","

            events.loc[events.id == event.id, '# Alerts'] = new_alert_count[:-1]
            email_counts.append(email_count[:-1])
        else:
            params = [event.id, event.Company]
            email_alerts = pd.read_sql_query(email_count_per_company_query, con, params=params)
            company_alerts = pd.read_sql_query(alert_count_per_company_query, con, params=params)

            alert_count = int(company_alerts.alert_count.values)
            total_alerts = int(events.loc[events.id == event.id, '# Alerts'].values)
            if alert_count != total_alerts: 
                # multi-company event, but a company filter must has been applied
                # update alert column to only alerts associated to the company
                events.loc[events.id == event.id, '# Alerts'] = alert_count

            # all mailbox alerts will be unique phish
            mailbox_phish = email_alerts.loc[email_alerts.alert_type=='mailbox']
            mailbox_phish_count = len(mailbox_phish)
            unique_phish = list(set(mailbox_phish.value.values))

            # remove mailbox alerts and leave any other alerts with a message_id observable
            email_alerts = email_alerts[email_alerts.alert_type!='mailbox']
            for alert in email_alerts.itertuples():
                if alert.value not in unique_phish:
                    unique_phish.append(alert.value)
                    mailbox_phish_count += 1
            email_counts.append(mailbox_phish_count)

    events['# Emails'] = email_counts

def get_events_between_dates(start_date: datetime,
                             end_date: datetime,
                             con: pymysql.connections.Connection,
                             event_query: str =EVENT_DB_QUERY,
                             selected_companies: list =[],
                            ) -> pd.DataFrame:
    """Query the database for all ACE events between two dates.

    Query the ACE database using the passed `db` connection using the `event_query`.
    If there are selected_companies, only return events associated to those companies,
    else all events are selected.

    Args:
        start_date: Get events created on or after this datetime.
        end_date: Get events created on or before this datetime.
        con: a pymysql database connectable
        event_query: The str SQL database query to get all events.
        selected_companies: A list of companies to select events for, by id.
          If the list is empty, all events are selected.

    Returns:
        A pd.DataFrame of the events.
    """

    # apply company selection by name
    company_ids = []
    if selected_companies:
        cursor = con.cursor()
        cursor.execute("select * from company")
        for c_id,c_name in cursor.fetchall():
            if c_name in selected_companies:
                company_ids.append(c_id)

    event_query = event_query.format(' AND ' if company_ids else '', '( ' + ' OR '.join(['company.id=%s' for company in selected_companies]) +') ' if company_ids else '')

    params = [start_date.strftime('%Y-%m-%d %H:%M:%S'),
              end_date.strftime('%Y-%m-%d %H:%M:%S')]
    params.extend(company_ids)

    events = pd.read_sql_query(event_query, con, params=params)
    events.set_index('Date', inplace=True)
    events.name = "Events"

    return events

def get_incidents_from_events(events: pd.DataFrame, incident_dispositions=INCIDENT_DISPOSITIONS) -> pd.DataFrame:
    """Get all incidents from events that have incident dispositions.

    If an event has an incident_disposition then it's an incident. Pull out
    a copy of those events and return in an incident pd.DataFrame.

    Args:
        events: A pd.DataFrame of ACE events
        incident_dispositions: The list of dispositions that mean an event was an incident.

    Returns:
        A pd.DataFrame of incidents.
    """

    incidents = events[events.Disposition.isin(incident_dispositions)]
    incidents = incidents.drop(columns=['id'])
    incidents.name = "Incidents"
    return incidents

def event_count_by_disposition(events: pd.DataFrame) -> dict:
    """Count event totals per disposition"""

    events.set_index('Disposition', inplace=True)
    dispositions = events.index.get_level_values('Disposition').unique()

    dispo_data = {}
    for dispo in dispositions:
        dispo_data[dispo] = len(events.loc[dispo])

    return dispo_data
    dispo_df = pd.DataFrame(data=dispo_data, columns=dispositions)
    return dispo_df

def count_event_dispositions_by_time_period(time_key: str,
                                            events: pd.DataFrame,
                                            dispositions=EVENT_DISPOSITIONS) -> pd.DataFrame:
    """Count events by their disposition over time.

       NOTE: events like EVENT_COUNT_TIME_DB_QUERY.
    """
    time_keys = events.index.get_level_values(time_key).unique()

    data_map = {}
    for dispo in dispositions:
        data_map[dispo] = {}
        for time_period in time_keys:
            data_map[dispo][time_period] = 0

            # all events during the time_period
            period_df = events.loc[time_period]

            if isinstance(period_df, pd.Series):
                # only a sinlge instance results in Series
                period_df = pd.DataFrame([period_df])

            period_dispo_data = event_count_by_disposition(period_df)

            if dispo in period_dispo_data:
                data_map[dispo][time_period] = period_dispo_data[dispo]

    return pd.DataFrame(data=data_map)
